import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange, repeat, einsum
import math


class Base(nn.Module):
    def __init__(self, divide=1.0, learned=False):
        super().__init__()
        self.learned = learned
        if not learned:
            self.div = divide
        else:
            self.div = nn.Parameter(torch.tensor(math.log(divide)))

        self.forward = self.mod_x

    def mod_x(self, x, pos=None, **kwargs):
        return x

    def mod_qkv(self, q, k, v, pos=None, **kwargs):
        return q, k, v

    def mod_mask(self, mask, q, k, v, pos=None, **kwargs):
        return mask

    def apply_pos_division(self, x):
        if not self.learned:
            return x / self.div
        else:
            return x / self.div.exp().clamp(min=1e-8)

    @staticmethod
    def default_pos_x(x):
        return torch.arange(x.size(-2), device=x.device).expand(*x.shape[:-1])


class Sinusoidal(Base):
    def __init__(self, dim, divide=1.0, learned_div=False):
        """Sinusoidal positional embedding as in Vaswani et al. (2017)
        Supports specifying positions, masking, and division of positional range.

        Args:
            dim (int): Hidden size of the embeddings
            divide (float, optional): divide positions by this factor, useful for large (or small) numerical ranges. Defaults to 1.0.
            learned_div (bool, optional): Whether to learn the division factor. Defaults to False.
        """
        super().__init__(divide=divide, learned=learned_div)
        self.dim = dim
        inv_freq = 1 / (10000 ** (torch.arange(0.0, dim, 2) / dim))
        self.register_buffer("inv_freq", inv_freq)

    def mod_x(self, x, pos=None, mask=None, **kwargs):
        """
        Args:
            x (torch.tensor): (B,*,L,H)
            pos (torch.tensor, optional): (B,*,L). Defaults to None for computing positions from 0 to L-1
            mask (torch.tensor, optional): (B,*,L). A boolean mask to indicate where positions should not have positional encodings added.
                Defaults to None for no masking
            **kwargs: ignored. Kept for compatibility.
            
        Returns:
            torch.tensor: x with added sinusoidal embeddings.
        """
        shp = list(x.shape)
        l = shp[-2]
        assert shp[-1] == self.dim

        pos_emb = self.default_pos_x(x).to(self.inv_freq) if pos is None else pos
        shp[-2] = pos_emb.shape[-1]

        sinusoid_inp = self.apply_pos_division(pos_emb.unsqueeze(-1) * self.inv_freq)
        pos_emb = (
            torch.stack((sinusoid_inp.sin(), sinusoid_inp.cos()), dim=-1)
            .view(*shp)
            .to(x.dtype)
        )

        if mask is not None:
            assert mask.dtype == torch.bool
            mask = F.pad(mask, (pos_emb.shape[-2]-mask.shape[-1], 0), value=True)
            pos_emb = pos_emb * mask.unsqueeze(-1)

        if l != shp[-2]:
            pos_emb = F.pad(pos_emb, (0, 0, l-shp[-2], 0))

        return x + pos_emb


class LearnedVocab(Base):
    def __init__(self, dim, max_seq_len):
        """Learned vocab as in Devlin et al. (2018)
        Supports specifying positions.
        Only works for discrete positional indices.

        Args:
            dim (int): Hidden size of the embeddings
            max_seq_len (int): maximum sequence length or vocab size of the learned embeddings
        """
        super().__init__()
        self.emb = nn.Embedding(max_seq_len, dim)

    def mod_x(self, x, pos=None, mask = None, **kwargs):
        """
        Args:
            x (torch.tensor): (B,*,L,H)
            pos (torch.tensor, optional): (B,*,L). Defaults to None for computing positions from 0 to L-1
            mask (torch.tensor, optional): (B,*,L). A boolean mask to indicate where positions should not have positional encodings added.
                Defaults to None for no masking
            **kwargs: ignored. Kept for compatibility.

        Returns:
            torch.tensor: x with added learned embeddings.
        """
        shp = list(x.shape)
        assert shp[-1] == self.emb.embedding_dim
        pos = self.default_pos_x(x).long() if pos is None else pos

        pos_emb = self.emb(pos)

        if mask is not None:
            assert mask.dtype == torch.bool
            mask = F.pad(mask, (pos_emb.shape[-2]-mask.shape[-1], 0), value=True)
            pos_emb = pos_emb * mask.unsqueeze(-1)

        l = pos_emb.shape[-2]
        if l != shp[-2]:
            pos_emb = F.pad(pos_emb, (0, 0, shp[-2]-l, 0))

        return x + pos_emb


class LearnedContinuous(Base):
    def __init__(self, dim, depth=3, norm=False, divide=1.0, learned_div=False):
        """Learned embeddings with a continuity between absolute positional indices, as learned by a series of linear layers.
        Supports specifying positions, and division of positional range.

        Args:
            dim (int): Hidden size of the embeddings
            depth (int, optional): number of hidden layers in the positional embedding network. Defaults to 3
            norm (bool, optional): Whether to use layernorms in the MLP. Defaults to False
            divide (float, optional): divide positions by this factor, useful for large (or small) numerical ranges. Defaults to 1.0.
            learned_div (bool, optional): Whether to learn the division factor. Defaults to False.
        """
        super().__init__(divide=divide, learned=learned_div)

        self.mlp = nn.ModuleList([])
        self.mlp.append(
            nn.Sequential(
                nn.Linear(1, dim),
                nn.LayerNorm(dim) if norm else nn.Identity(),
            )
        )

        for _ in range(depth - 1):
            self.mlp.append(
                nn.Sequential(
                    nn.SiLU(),
                    nn.Linear(dim, dim),
                    nn.LayerNorm(dim) if norm else nn.Identity(),
                )
            )

    def mod_x(self, x, pos=None, mask=None, **kwargs):
        """
        Args:
            x (torch.tensor): (B,*,L,H)
            pos (torch.tensor, optional): (B,*,L). Defaults to None for computing positions from 0 to L-1
            mask (torch.tensor, optional): (B,*,L). A boolean mask to indicate where positions should not have positional encodings added.
                Defaults to None for no masking
            **kwargs: ignored. Kept for compatibility.
            
        Returns:
            torch.tensor: x with added learned embeddings.
        """
        shp = list(x.shape)

        pos = self.default_pos_x(x).to(x) if pos is None else pos

        pos_emb = self.apply_pos_division(pos).unsqueeze(-1)
        for layer in self.mlp:
            pos_emb = layer(pos_emb)

        if mask is not None:
            assert mask.dtype == torch.bool
            mask = F.pad(mask, (pos_emb.shape[-2]-mask.shape[-1], 0), value=True)
            pos_emb = pos_emb * mask.unsqueeze(-1)

        l = pos_emb.shape[-2]
        if l != shp[-2]:
            pos_emb = F.pad(pos_emb, (0, 0, shp[-2]-l, 0))

        return x + pos_emb


class Rotary(Base):
    def __init__(self, head_dim, n_dims=None, divide=1.0, learned_div=False):
        """Rotary embedding as in RoFormer / RoPE
        Supports specifying positions and division of positional range.

        Args:
            head_dim (int): Hidden dimensions per head
            n_dims (int, optional):
                number of dimensions (per head) to apply rotations on.
                Can be used to control how strong the positional bias should be.
                Defaults to None to use all dims
            divide (float, optional): divide positions by this factor, useful for large (or small) numerical ranges. Defaults to 1.0.
            learned_div (bool, optional): Whether to learn the division factor. Defaults to False.
        """
        super().__init__(divide=divide, learned=learned_div)

        thetas = 1.0 / (10000 ** (torch.arange(0, head_dim, 2).float() / head_dim))
        if n_dims is not None:
            assert n_dims % 2 == 0
            thetas[(n_dims // 2) :] = 0
        self.register_buffer("thetas", thetas)

    def mod_qkv(
        self,
        q,
        k,
        v,
        pos_q_k=None,
        pos_q=None,
        pos_k=None,
        self_attn_mode=True,
        **kwargs,
    ):
        """
        Args:
            q (torch.tensor): (B,*,L1,NH,H)
            k (torch.tensor): (B,*,L2,NH,H)
            v (torch.tensor): (B,*,L2,NH,H)
            pos_q_k (torch.tensor, optional): (B,*,L). Positions of q and k in self attention mode. Requires L1 = L2.
                Defaults to None for computing positions from 0 to L-1
            pos_q (torch.tensor, optional): (B,*,L1). Positions of q in cross attention mode.
                Defaults to None for computing positions from 0 to L1-1
            pos_k (torch.tensor, optional): (B,*,L2). Positions of k in cross attention mode.
                Defaults to None for computing positions from 0 to L2-1
            self_attn_mode (bool, optional): Whether to use the same positions for q and k (pos_q_k),
                or use different positions for each (pos_q) and (pos_k).
                Defaults to True.
            **kwargs: ignored. Kept for compatibility.
        Returns:
            q, k, v (all torch.tensor)
        """
        if self_attn_mode:
            assert q.shape[-3] == k.shape[-3]
            pos_q_k = self.default_pos_x(q[..., 0, :]) if pos_q_k is None else pos_q_k
            sin, cos = self.get_rotations(pos_q_k)

            l_sin, l_q = sin.shape[-3], q.shape[-3]
            if l_sin != l_q:
                sin, cos = map(lambda t: F.pad(t, (0, 0, 0, 0, l_q-l_sin, 0), (sin, cos)))

            q = q * cos.to(q) + self.rotate_every_two(q) * sin.to(q)
            k = k * cos.to(k) + self.rotate_every_two(k) * sin.to(k)
        else:
            pos_q = self.default_pos_x(q[..., 0, :]) if pos_q is None else pos_q
            pos_k = self.default_pos_x(k[..., 0, :]) if pos_k is None else pos_k
            sin_q, cos_q = self.get_rotations(pos_q)
            sin_k, cos_k = self.get_rotations(pos_k)

            l_sin, l_q = sin_q.shape[-3], q.shape[-3]
            if l_sin != l_q:
                sin_q, cos_q = map(lambda t: F.pad(t, (0, 0, 0, 0, l_q-l_sin, 0), (sin_q, cos_q)))
            l_sin, l_k = sin_k.shape[-3], k.shape[-3]
            if l_sin != l_k:
                sin_k, cos_k = map(lambda t: F.pad(t, (0, 0, 0, 0, l_k-l_sin, 0), (sin_k, cos_k)))

            q = q * cos_q.to(q) + self.rotate_every_two(q) * sin_q.to(q)
            k = k * cos_k.to(k) + self.rotate_every_two(k) * sin_k.to(k)

        return q, k, v

    def get_rotations(self, pos):
        mthetas = (
            self.apply_pos_division(pos).to(self.thetas)[..., None] * self.thetas
        )  # B, *, h
        sin, cos = map(
            lambda t: repeat(t, "b ... h  -> b ... (h j)", j=2),
            (mthetas.sin(), mthetas.cos()),
        )
        sin, cos = map(lambda t: t.unsqueeze(-2), (sin, cos))
        return sin, cos

    @staticmethod
    def rotate_every_two(x):
        x = x.clone()
        x = rearrange(x, "... (d j) -> ... d j", j=2)
        x1, x2 = x.unbind(dim=-1)
        x = torch.stack((-x2, x1), dim=-1)
        return rearrange(x, "... d j -> ... (d j)")


class ALiBi(Base):
    def __init__(
        self, n_heads, use_n_heads=None, asymmetric=False, divide=1.0, learned_div=False
    ):
        """Attention with linear biases as in Press et al. 2022
        Supports specifying positions, division of positional range, using only a fraction of the heads, and asymmetric bias for bidirectional cases.


        Args:
            n_heads (int): Number of heads
            use_n_heads (int, optional):
                Number of heads to use
                Can be used to control how strong the positional bias should be.
                Defaults to None to use all heads
            assymetric (bool, optional):
                Whether to use assymetric positional biases to differentiate between negative or positive relative positions.
                Implemented according to solution 3 proposed here https://github.com/ofirpress/attention_with_linear_biases/issues/5.
                Defaults to False.
            divide (float, optional): divide positions by this factor, useful for large (or small) numerical ranges. Defaults to 1.0.
            learned_div (bool, optional): Whether to learn the division factor. Defaults to False.
        """
        super().__init__(divide=divide, learned=learned_div)

        self.asymmetric = asymmetric
        use_n_heads = n_heads if use_n_heads is None else use_n_heads
        if not asymmetric:
            slopes = torch.tensor(
                self._get_slopes(use_n_heads) + [0] * (n_heads - use_n_heads)
            )
        else:
            slopes = self._get_slopes(use_n_heads)
            slopes = [slopes[(i // 2) * 2] for i in range((len(slopes)))]
            slopes = torch.tensor(slopes + [0] * (n_heads - use_n_heads))
        slopes = rearrange(slopes, "h -> h 1 1")
        self.register_buffer("slopes", slopes)

        self.div = divide

    def mod_mask(self, mask, q, k, v, pos_q_k=None, pos_q=None, pos_k=None, **kwargs):
        """
        Args:
            mask (torch.tensor): B, *, NH, L1, L2, optional, eg B, NH, L1, L2, or for multi-dim: B, S, NH, L1, L2
            q (torch.tensor): (B,*,L1,NH,H)
            k (torch.tensor): (B,*,L2,NH,H)
            v (torch.tensor): (B,*,L2,NH,H)
            pos_q_k (torch.tensor, optional): (B, *, L). Positions of q and k in self attention mode. Requires L1 = L2 =L.
                Defaults to None for computing positions from 0 to L-1.
            pos_q (torch.tensor, optional): (B, *, L) eg (B, L1). Positions of q in cross attention mode.
                Ignored if pos_q_k is specified
                Defaults to None for computing positions from 0 to L1-1.
            pos_k (torch.tensor, optional): (B, *, L) eg (B, L2). Positions of k in cross attention mode.
                Ignored if pos_q_k is specified
                Defaults to None for computing positions from 0 to L2-1.
            **kwargs: ignored. Kept for compatibility.
        Returns:
            mask (torch.tensor): (B, *, NH, L1, L2)
        """

        if pos_q_k is not None:
            pos_q = pos_k = pos_q_k

        elif (pos_q is None) and (pos_k is None):
            pos_q = self.default_pos_x(q[..., 0, :]) if pos_q is None else pos_q
            pos_k = self.default_pos_x(k[..., 0, :]) if pos_k is None else pos_k

        relative_pos = pos_q[..., :, None] - pos_k[..., None, :]
        bias = self.apply_pos_division(relative_pos.unsqueeze(-3) * self.slopes)

        if self.asymmetric:
            bias = self.asymmetric_bias(bias)
        bias = torch.abs(bias)

        l1, l2 = q.shape[-3], k.shape[-3]
        mask_l1, mask_l2 = mask.shape[-2], mask.shape[-1]
        if (l1 != mask_l1) or (l2 != mask_l2):
            mask = F.pad(mask, (l2-mask_l2, 0, l1-mask_l1, 0))

        return (0 if mask is None else mask) - bias

    @staticmethod
    def _get_slopes(heads):
        def get_slopes_power_of_2(n):
            start = 2 ** (-(2 ** -(math.log2(n) - 3)))
            ratio = start
            return [start * ratio**i for i in range(n)]

        if math.log2(heads).is_integer():
            return get_slopes_power_of_2(heads)

        closest_power_of_2 = 2 ** math.floor(math.log2(heads))
        return (
            get_slopes_power_of_2(closest_power_of_2)
            + get_slopes_power_of_2(2 * closest_power_of_2)[0::2][
                : heads - closest_power_of_2
            ]
        )

    @staticmethod
    def asymmetric_bias(bias):
        bias[:, ::2][bias[:, ::2] > 0] = torch.finfo(bias.dtype).max
        bias[:, 1::2][bias[:, 1::2] < 0] = torch.finfo(bias.dtype).max
        return bias


class DPB(Base):
    def __init__(
        self, dim, n_heads, depth=3, norm=False, divide=1.0, learned_div=False
    ):
        """Dynamic positional bias.
        Computes a bias to every element of the attention matrix based on their relative position via a parameterized MLP
        Described in https://arxiv.org/abs/2108.00154, https://arxiv.org/abs/2111.09883,
        https://github.com/lucidrains/x-transformers#dynamic-positional-bias
        Supports specifying positions and division of positional range


        Args:
            dim (int): number of hidden dimensions of the MLP
            n_heads (int): Number of heads in the model
            depth (int, optional): number of hidden layers in the positional embedding network. Defaults to 3
            norm (bool, optional): Whether to use layernorms in the MLP. Defaults to False
            divide (float, optional): divide positions by this factor, useful for large (or small) numerical ranges. Defaults to 1.0.
            learned_div (bool, optional): Whether to learn the division factor. Defaults to False.
        """
        super().__init__(divide=divide, learned=learned_div)
        self.mlp = nn.ModuleList([])
        self.mlp.append(
            nn.Sequential(
                nn.Linear(1, dim),
                nn.LayerNorm(dim) if norm else nn.Identity(),
                nn.SiLU(),
            )
        )

        for _ in range(depth - 1):
            self.mlp.append(
                nn.Sequential(
                    nn.Linear(dim, dim),
                    nn.LayerNorm(dim) if norm else nn.Identity(),
                    nn.SiLU(),
                )
            )

        self.mlp.append(nn.Linear(dim, n_heads))

        self.div = divide

    def mod_mask(self, mask, q, k, v, pos_q_k=None, pos_q=None, pos_k=None, **kwargs):
        """
        Args:
            mask (torch.tensor): B, *, NH, L1, L2, optional, eg B, NH, L1, L2, or for multi-dim: B, S, NH, L1, L2
            q (torch.tensor): (B,*,L1,NH,H)
            k (torch.tensor): (B,*,L2,NH,H)
            v (torch.tensor): (B,*,L2,NH,H)
            pos_q_k (torch.tensor, optional): (B, *,L). Positions of q and k in self attention mode. Requires L1 = L2 = L.
                Defaults to None for computing positions from 0 to L-1
            pos_q (torch.tensor, optional): (B, *,L1). Positions of q in cross attention mode.
                Defaults to None for computing positions from 0 to L1-1
            pos_k (torch.tensor, optional): (B, *,L2). Positions of k in cross attention mode.
                Defaults to None for computing positions from 0 to L2-1
            **kwargs: ignored. Kept for compatibility.
        Returns:
            mask (torch.tensor): (B, *, NH, L1, L2)
        """
        if pos_q_k is not None:
            pos_q = pos_k = pos_q_k
        elif (pos_q is None) and (pos_k is None):
            pos_q = self.default_pos_x(q[..., 0, :]) if pos_q is None else pos_q
            pos_k = self.default_pos_x(k[..., 0, :]) if pos_k is None else pos_k

        bias = self.apply_pos_division(pos_q[..., :, None] - pos_k[..., None, :])[
            ..., None
        ].to(q)

        for layer in self.mlp:
            bias = layer(bias)
        bias = bias.transpose(-1, -2).transpose(-2, -3).to(q)

        l1, l2 = q.shape[-3], k.shape[-3]
        mask_l1, mask_l2 = mask.shape[-2], mask.shape[-1]
        if (l1 != mask_l1) or (l2 != mask_l2):
            mask = F.pad(mask, (l2-mask_l2, 0, l1-mask_l1, 0))

        return (0 if mask is None else mask) - bias


class XL(Base):
    def __init__(self, dim, n_heads, divide=1, learned_div=False):
        """Relative positional biases as in Transformer-XL (Dai et al 2019)
        Supports specifying positions and division of positional range

        Args:
            dim (int): number of hidden of x (total, not per head)
            n_heads (int): Number of heads in the model
            divide (float, optional): divide positions by this factor, useful for large (or small) numerical ranges. Defaults to 1.0.
            learned_div (bool, optional): Whether to learn the division factor. Defaults to False.
        """
        super().__init__(divide=divide, learned=learned_div)
        self.embed_lin = nn.Linear(dim, dim)
        self.bias_r_w = self._init_bias(
            nn.Parameter(torch.Tensor(n_heads, dim // n_heads))
        )
        self.bias_r_r = self._init_bias(
            nn.Parameter(torch.Tensor(n_heads, dim // n_heads))
        )
        self.n_heads = n_heads
        self.dim = dim
        inv_freq = 1 / (10000 ** (torch.arange(0.0, dim, 2) / dim))
        self.register_buffer("inv_freq", inv_freq)
        self.div = divide

    def mod_qkv(self, q, k, v, **kwargs):
        """
        Args:
            q (torch.tensor): (B,*,L1,NH,H)
            k (torch.tensor): (B,*,L2,NH,H)
            v (torch.tensor): (B,*,L2,NH,H)
        Returns:
            q, k, v (all torch.tensor)
        """
        return q + self.bias_r_w, k, v

    def mod_mask(self, mask, q, k, v, pos_q_k=None, pos_q=None, pos_k=None, **kwargs):
        """
        Args:
            mask (torch.tensor): B, *, NH, L1, L2, optional, eg B, NH, L1, L2, or for multi-dim: B, S, NH, L1, L2
            q (torch.tensor): (B,*,L1,NH,H)
            k (torch.tensor): (B,*,L2,NH,H)
            v (torch.tensor): (B,*,L2,NH,H)
            pos_q_k (torch.tensor, optional): (B, *, L). Positions of q and k in self attention mode. Requires L1 = L2 = L.
                Defaults to None for computing positions from 0 to L-1.
            pos_q (torch.tensor, optional): (B, *, L1). Positions of q in cross attention mode.
                Defaults to None for computing positions from 0 to L1-1
            pos_k (torch.tensor, optional): (B, *, L2). Positions of k in cross attention mode.
                Defaults to None for computing positions from 0 to L2-1
            **kwargs: ignored. Kept for compatibility.
        Returns:
            mask (torch.tensor): (B, *, NH, L1, L2)
        """
        if pos_q_k is not None:
            pos_q = pos_k = pos_q_k
        elif (pos_q is None) and (pos_k is None):
            pos_q = self.default_pos_x(q[..., 0, :]) if pos_q is None else pos_q
            pos_k = self.default_pos_x(k[..., 0, :]) if pos_k is None else pos_k

        relative_pos = (pos_q[..., :, None] - pos_k[..., None, :]).to(
            self.inv_freq.dtype
        )

        sinusoid_inp = self.apply_pos_division(relative_pos[..., None] * self.inv_freq)
        pos_emb = torch.stack((sinusoid_inp.sin(), sinusoid_inp.cos()), dim=-1)

        pos_emb = rearrange(pos_emb, "... d j -> ... (d j)")

        r = rearrange(
            self.embed_lin(pos_emb),
            "... (nh h) -> ... nh h",
            nh=self.n_heads,
            h=q.shape[-1],
        )

        l1, l2 = q.shape[-3], k.shape[-3]
        r_l1, r_l2 = r.shape[-4], r.shape[-3]
        if (l1 != r_l1) or (l2 != r_l2):
            r = F.pad(r, (0, 0, 0, 0, l2-r_l2, 0, l1-r_l1, 0))

        q_r = q + self.bias_r_w
        BD = einsum(q_r, r, "... q n h, ... q k n h -> ... n q k")
        return (0 if mask is None else mask) + BD

    def _init_bias(self, bias):
        bound = 1 / bias.size(1) ** 0.5
        return nn.init.uniform_(bias, -bound, bound)



import torch
from torch import nn, einsum
from einops import rearrange, repeat
import torch.nn.functional as F
import math
from bio_attention import positional


def compl_mod(m, n):
    return int(n * math.ceil(m / n) - m)


class Attention(torch.nn.Module):
    def __init__(
        self,
        dropout=0.0,
        enable_math=True,
        enable_flash=True,
        enable_mem_efficient=True,
        **kwargs,
    ):
        super().__init__()
        self.dropout = dropout
        self.context_manager = {
            "enable_math": enable_math,
            "enable_flash": enable_flash,
            "enable_mem_efficient": enable_mem_efficient,
        }
        self.use_context_manager = not all(
            [enable_math, enable_flash, enable_mem_efficient]
        )

    def forward(self, q, k, v, mask=None, causal=False):
        """Default attention layer

        Parameters
        ----------
        q : B, *, L1, NH, H
            queries
        k : B, *, L2, NH, H
            keys
        v : B, *, L2, NH, H
            values
        mask : B, *, NH, L1, L2, optional
            _description_, by default None
        causal : bool, optional
            whether to do causal attention or not, by default False

        Returns
        -------
        _type_
            _description_
        """
        q_shape = q.shape
        q, k, v = map(
            lambda t: rearrange(t, "b ... x n h -> (b ...) x n h").permute(0, 2, 1, 3),
            (q, k, v),
        ) # (B...), NH, L, H
        if mask is not None:
            mask = rearrange(mask, "b ... n q k -> (b ...) n q k") # (B...), NH, L1, L2

        if causal:
            causal_mask = torch.ones(q.shape[-2], k.shape[-2], dtype=torch.bool)
            causal_mask = causal_mask.triu(diagonal=0).expand(q.shape[0], q.shape[1], -1, -1)
            mask = (mask if mask is not None else (causal_mask).to(q)).masked_fill(causal_mask, -float('inf'))

        if self.use_context_manager:
            with torch.backends.cuda.sdp_kernel(**self.context_manager):
                return (
                    F.scaled_dot_product_attention(
                        q,
                        k,
                        v,
                        dropout_p=(self.dropout if self.training else 0),
                        attn_mask=mask,
                        is_causal=False,
                    )
                    .permute(0, 2, 1, 3)
                    .view(*q_shape)
                )
        else:
            return (
                F.scaled_dot_product_attention(
                    q,
                    k,
                    v,
                    dropout_p=(self.dropout if self.training else 0),
                    attn_mask=mask,
                    is_causal=False,
                )
                .permute(0, 2, 1, 3)
                .view(*q_shape)
            )


class RandomAttention(nn.Module):
    def __init__(
        self,
        n_random_keys=64,
        dropout=0.2,
        materialize_full=False,
        **kwargs,
    ):
        super().__init__()
        if not materialize_full:
            self.softmax = nn.Softmax(dim=-2)
            self.dropout = nn.Dropout(dropout)

            self.forward = self.forward_indexed
        else:
            self.dropout = dropout
            self.forward = self.forward_naive
        self.n = n_random_keys

    def forward_indexed(self, q, k, v, mask=None, causal=False):
        """random attention layer

        NOTE: causal attention will erase input masks

        Parameters
        ----------
        q : B, *, L1, NH, H
            queries
        k : B, *, L2, NH, H
            keys
        v : B, *, L2, NH, H
            values
        mask : B, *, NH, L1, L2, optional
            _description_, by default None
        causal : bool, optional
            whether to do causal attention or not, by default False

        Returns
        -------
        _type_
            _description_
        """
        assert not (causal and (mask is not None))

        q_shape = q.shape
        q, k, v = map(lambda t: rearrange(t, "b ... x n h -> (b ...) x n h"), (q, k, v))
        if mask is not None:
            mask = rearrange(mask, "b ... n q k -> (b ...) n q k")

        b, s2, nh, h = k.shape
        s1 = q.shape[1]

        mask = (
            torch.ones(b, nh, s1, s2, dtype=torch.bool).tril(diagonal=0)
            if causal
            else mask
        )
        if mask is not None:
            mask = (
                (~mask).float().masked_fill(~mask, -float("inf"))
                if mask.dtype == torch.bool
                else mask
            )
            mask = mask.to(q)

        q = q * (h**-0.5)

        indices_select = torch.randint(0, s1, (b, s2, self.n)).to(q.device)
        indexer = torch.arange(b).view(b, 1, 1)
        if mask is not None:
            mask = mask.permute(0, 2, 3, 1)[
                indexer, torch.arange(s1)[None, :, None], indices_select
            ].permute(0, 3, 1, 2)

        k = k[indexer, indices_select]
        v = v[indexer, indices_select]

        A = einsum("b q n h, b q k n h -> b q k n", q, k) + (
            0 if mask is None else mask.permute(0, 2, 3, 1)
        )
        A = self.softmax(A)

        A = self.dropout(A)
        z = einsum("b q k n, b q k n h -> b q n h", A, v)

        return z.view(*q_shape)

    def forward_naive(self, q, k, v, mask=None, causal=False):
        """NOTE: Is incompatible with causal attention.
        NOTE: Is incompatible with input masks attention.

        Args:
            q (_type_): _description_
            k (_type_): _description_
            v (_type_): _description_
            mask (_type_, optional): _description_. Defaults to None.
            causal (bool, optional): _description_. Defaults to False.

        Returns:
            _type_: _description_
        """
        q_shape = q.shape
        q, k, v = map(
            lambda t: rearrange(t, "b ... x n h -> (b ...) x n h").permute(0, 2, 1, 3),
            (q, k, v),
        )
        if mask is not None:
            mask = rearrange(mask, "b ... n q k -> (b ...) n q k")

        b, nh, s2, h = k.shape
        s1 = q.shape[-2]

        mask = (
            (torch.rand(b, 1, s1, s2) < (self.n / s2))
            .expand(-1, nh, -1, -1)
            .to(k.device)
        )

        return (
            F.scaled_dot_product_attention(
                q,
                k,
                v,
                dropout_p=(self.dropout if self.training else 0),
                attn_mask=mask,
                is_causal=causal,
            )
            .permute(0, 2, 1, 3)
            .view(*q_shape)
        )


class WindowAttention(nn.Module):
    def __init__(
        self,
        window=5,
        dropout=0.1,
        materialize_full=False,
        **kwargs,
    ):
        super().__init__()
        assert window % 2 == 1, "Window size should be an odd integer."

        self.w = int((window - 1) / 2)

        if not materialize_full:
            self.softmax = nn.Softmax(dim=-1)
            self.dropout = nn.Dropout(dropout)
            self.k_ch = window * 2
            self.q_ch = window + 1

            u = torch.triu(torch.full((self.q_ch, self.k_ch), True))
            self.mask = ~torch.logical_and(u, torch.flip(u, [0, 1]))
            self.mask_k_left = torch.clone(self.mask)
            self.mask_k_left[:, : self.w] = True
            self.forward = self.forward_sliced

        else:
            self.dropout = dropout
            self.forward = self.forward_naive

    def forward_sliced(self, q, k, v, mask=None, causal=False):
        """windowed attention layer

        Parameters
        ----------
        q : B, L1, NH, H
            queries
        k : B, L2, NH, H
            keys
        v : B, L2, NH, H
            values
        mask : B, NH, L1, L2, optional
            _description_, by default None
        causal : bool, optional
            whether to do causal attention or not, by default False

        Returns
        -------
        _type_
            _description_

        NOTE: incompatible with user-defined masks
        """
        assert mask is None

        assert k.shape[-3] == q.shape[-3], "q and k should have same input length."

        q_shape = q.shape
        q, k, v = map(lambda t: rearrange(t, "b ... x n h -> (b ...) x n h"), (q, k, v))

        b, s, nh, h = k.shape

        q = q * (h**-0.5)

        q_pad = compl_mod(s, self.q_ch)
        k_pad = compl_mod((s + self.w * 2) - self.k_ch, self.q_ch)

        q = F.pad(q, (0,) * 5 + (q_pad,)).unfold(1, self.q_ch, self.q_ch)
        k = F.pad(k, (0,) * 4 + (self.w, self.w + k_pad)).unfold(
            1, self.k_ch, self.q_ch
        )
        v = F.pad(v, (0,) * 4 + (self.w, self.w + k_pad)).unfold(
            1, self.k_ch, self.q_ch
        )

        A = einsum("b c n h q, b c n h k -> b n c q k ", q, k)

        mask_value = -torch.finfo(A.dtype).max
        mask_k_right = torch.clone(self.mask.to(A.device))
        mask_k_right[:, -(self.w + k_pad) :] = True
        if q.shape[1] > 1:
            mask = torch.stack(
                [self.mask_k_left.to(A.device)]
                + [self.mask.to(A.device)] * (q.shape[1] - 2)
                + [mask_k_right]
            )
        else:
            mask = torch.logical_or(
                self.mask_k_left.to(A.device), mask_k_right
            ).unsqueeze(0)
        if causal:
            mask = ~(~mask).tril(diagonal=self.w)

        A.masked_fill_(mask, mask_value)
        A = self.softmax(A)
        A = self.dropout(A)

        z = einsum("b n c q k, b c n h k -> b n c q h", A, v)
        z = z.view(b, nh, -1, h)[:, :, :s].permute(0, 2, 1, 3).view(*q_shape)

        return z

    def forward_naive(self, q, k, v, mask=None, causal=False):
        assert mask is None
        assert k.shape[-3] == q.shape[-3], "q and k should have same input length."

        q_shape = q.shape
        q, k, v = map(
            lambda t: rearrange(t, "b ... x n h -> (b ...) x n h").permute(0, 2, 1, 3),
            (q, k, v),
        )

        b, nh, s2, h = k.shape
        s1 = q.shape[-2]

        u = torch.triu(torch.full((s1, s2), True), diagonal=-self.w)
        mask = (
            torch.logical_and(u, torch.flip(u, [0, 1]))
            .expand(b, nh, -1, -1)
            .to(k.device)
        )
        if causal:
            mask = torch.tril(mask)
        return (
            F.scaled_dot_product_attention(
                q,
                k,
                v,
                dropout_p=(self.dropout if self.training else 0),
                attn_mask=mask,
                is_causal=False,
            )
            .permute(0, 2, 1, 3)
            .view(*q_shape)
        )


class AttnLayer(nn.Module):
    def __init__(self, dim, attn, nh=4, plugin=None):
        super().__init__()
        assert dim % nh == 0, "dim should be divisible by number of heads"

        self.lin = nn.Linear(dim, 3 * dim)
        self.attn = attn
        self.nh = nh

        self.plugin = plugin if plugin is not None else positional.Base()

    def forward(self, x, pos=None, mask=None, causal=False, **mod_kwargs):
        """
        MASK = is used by mod_x - type positional embeddings if it is of bool type and shape B, *, L
        MASK can be (B,*,L), (B,*,L,L), or (B, * NH, L, L).
        x = (B, *, L, H)
        """
        if (mask is not None) and (mask.dtype == torch.bool):
            mask = F.pad(mask,(x.shape[-2] - mask.shape[-1], 0), value = True)

        use_mask_to_mod_x = False
        if mask is not None:
            use_mask_to_mod_x = (mask.dtype == torch.bool) and (mask.shape == x.shape[:-1])

        x = self.plugin.mod_x(x, pos=pos, mask = (mask if use_mask_to_mod_x else None), **mod_kwargs)
        q, k, v = torch.split(self.lin(x), x.size(-1), dim=-1)
        q, k, v = map(
            lambda t: rearrange(t, "... (n h) -> ... n h", n=self.nh), (q, k, v)
        ) # B, *, L, NH, H
        if mask is not None:
            if q.ndim - mask.ndim == 2: #B, *, L -> B, *, L, L
                mask = repeat(mask, '... l -> ... (l2) l', l2=mask.shape[-1])
                
            if q.ndim - mask.ndim == 1: #B, *, L, L  -> B, *, NH, L, L
                mask = repeat(mask, '... l1 l2 -> ... nh l1 l2', nh=q.shape[-2])
            
            assert mask.shape[:-3] == q.shape[:-3]
            if mask.dtype == torch.bool:
                mask = (~mask).to(q).masked_fill(~mask, -float('inf'))

            mask = F.pad(mask, (k.shape[-3] - mask.shape[-1], 0, (q.shape[-3] - mask.shape[-2]), 0), value = 0)

        mod_kwargs |= {"pos_q_k": pos}

        q, k, v = self.plugin.mod_qkv(q, k, v, **mod_kwargs)

        mask = self.plugin.mod_mask(mask, q, k, v, **mod_kwargs)

        return rearrange(
            self.attn(q, k, v, mask=mask, causal=causal),
            "... n h -> ... (n h)",
            n=self.nh,
        )

class Residual(nn.Module):
    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    def forward(self, x, *args, **kwargs):
        return self.fn(x, *args, **kwargs) + x


class GLU(nn.Module):
    def __init__(self, dim, ff_dim, activation):
        super().__init__()
        self.act = activation
        self.proj = nn.Linear(dim, ff_dim * 2)

    def forward(self, x):
        x, gate = self.proj(x).chunk(2, dim=-1)
        return x * self.act(gate)


class TransformerLayer(nn.Module):
    def __init__(
        self, attn, dim, nh, plugin=None, dropout=0.2, glu_ff=True, activation="swish"
    ):
        super().__init__()
        if activation.lower() == "relu":
            act = nn.ReLU()
        elif activation.lower() == "gelu":
            act = nn.GELU()
        elif activation.lower() == "swish":
            act = nn.SiLU()

        self.norm = nn.LayerNorm(dim)
        self.attn = AttnLayer(dim, attn, nh=nh, plugin=plugin)

        project_in = (
            nn.Sequential(nn.Linear(dim, 4 * dim), act)
            if not glu_ff
            else GLU(dim, 4 * dim, act)
        )
        self.ff = Residual(
            nn.Sequential(
                nn.LayerNorm(dim),
                project_in,
                nn.Dropout(dropout),
                nn.Linear(4 * dim, dim),
            )
        )

    def forward(self, x, pos=None, mask=None, causal=False, **mod_kwargs):
        x = self.attn(self.norm(x), pos=pos, mask=mask, causal=causal, **mod_kwargs) + x
        return self.ff(x)


class Transformer(nn.Module):
    def __init__(
        self,
        depth,
        dim,
        nh,
        attentiontype="vanilla",
        attention_args={},
        plugintype="none",
        plugin_args={},
        only_apply_plugin_at_first=False,
        dropout=0.2,
        glu_ff=True,
        activation="swish",
    ):
        super().__init__()

        assert attentiontype in ["vanilla", "random", "window"]

        if attentiontype == "vanilla":
            attn_op = Attention(**attention_args)
        elif attentiontype == "random":
            attn_op = RandomAttention(**attention_args)
        elif attentiontype == "window":
            attn_op = WindowAttention(**attention_args)

        if plugintype == "none":
            plugin = positional.Base()
        elif plugintype == "sinusoidal":
            plugin = positional.Sinusoidal(**plugin_args)
        elif plugintype == "learned":
            plugin = positional.LearnedVocab(**plugin_args)
        elif plugintype == "learnedcont":
            plugin = positional.LearnedContinuous(**plugin_args)
        elif plugintype == "rotary":
            plugin = positional.Rotary(**plugin_args)
        elif plugintype == "ALiBi":
            plugin = positional.ALiBi(**plugin_args)
        elif plugintype == "DPB":
            plugin = positional.DPB(**plugin_args)
        elif plugintype == "XL":
            plugin = positional.XL(**plugin_args)

        layers = []
        layers.append(
            TransformerLayer(
                attn_op,
                dim,
                nh,
                plugin=plugin,
                dropout=dropout,
                glu_ff=glu_ff,
                activation=activation,
            )
        )
        
        for _ in range(depth - 1):
            layers.append(
                TransformerLayer(
                    attn_op,
                    dim,
                    nh,
                    plugin=(None if only_apply_plugin_at_first else plugin),
                    dropout=dropout,
                    glu_ff=glu_ff,
                    activation=activation,
                )
            )
    

        self.layers = nn.ModuleList(layers)

    def forward(self, x, pos=None, mask=None, causal=False, **mod_kwargs):
        for layer in self.layers:
            x = layer(x, pos=pos, mask=mask, causal=causal, **mod_kwargs)
        return x

class TransformerEncoder(Transformer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    def forward(self, x, pos=None, mask=None, **mod_kwargs):
        for layer in self.layers:
            x = layer(x, pos=pos, mask=mask, causal=False, **mod_kwargs)
        return x
    
class TransformerDecoder(Transformer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    def forward(self, x, pos=None, mask=None, **mod_kwargs):
        for layer in self.layers:
            x = layer(x, pos=pos, mask=mask, causal=True, **mod_kwargs)
        return x


class Aggregator(nn.Module):
    def __init__(self, method = "max"):
        super().__init__()
        assert method in ["mean", "max", "cls"]
        self.method = method

    def forward(self, x, mask = None):
        """
        X = B, *, L, H
        mask = B, *, L
        """
        if mask is not None:
            x = x * mask.unsqueeze(-1)
        if self.method == "mean":
            return x.sum(-2) / (mask.sum(-1, keepdim=True) if mask is not None else x.shape[-2])
        elif self.method == "max":
            return x.max(-2).values
        elif self.method == "cls":
            return x[..., 0, :]


import torch
from torch import nn


class DiscreteEmbedding(nn.Module):
    def __init__(
        self,
        num_embeddings,
        embedding_dim,
        cls=False,
        cls_axis=-2,
        init_cls_as_mask=False,
    ):
        """Embedding module for discrete data. Uses `nn.Embedding` with added support for masking tokens and cls tokens.
        Masking tokens are recognized as torch.nans. Therefore, input should be float dtype.
        WARNING: Because this module expects a float dtype for what should be integers, there is a limit to the vocab size you can effectively accurately represent.
        In practice, vocab sizes up to 1 million in size should be faithfully represented with float32. For float16, expect issues.

        Args:
            num_embeddings (int): number of discrete classes.
            embedding_dim (int): number of hidden dimensions of the embeddings
            cls (bool, optional): whether to include cls token. Defaults to True.
            cls_axis (int, optional): which axis in the input to add cls token to. Defaults to -1.
            init_cls_as_mask (bool, optional): initialize the cls token to be equal to the masking token. Defaults to False.
        """
        super().__init__()
        self.embedder = nn.Embedding(num_embeddings + 1, embedding_dim)

        if cls:
            self.cls = CLS(
                embedding_dim,
                cls_axis=cls_axis,
                init=(
                    None
                    if not init_cls_as_mask
                    else self.embedder.weight.data[0].clone()
                ),
            )
        else:
            self.cls = nn.Identity()

    def forward(self, x):
        x = x + 1
        x[torch.isnan(x)] = 0
        x = self.embedder(x.long())
        x = self.cls(x)
        return x


class ContinuousEmbedding(nn.Module):
    def __init__(
        self,
        embedding_dim,
        depth=1,
        norm=False,
        cls=False,
        cls_axis=-2,
        init_cls_as_bias=False,
        init_mask_as_bias=False,
        init_cls_as_mask=False,
    ):
        """Embedding module for continuous data. Uses `nn.Linear` with added support for masking tokens cls tokens.
        Masking tokens are recognized as torch.nans. Therefore, input should be float dtype.

        Args:
            embedding_dim (int): number of hidden dimensions of the embeddings
            cls (bool, optional): whether to include cls token. Defaults to True.
            cls_axis (int, optional): which axis in the input to add cls token to. Defaults to -1.
            init_cls_as_bias (bool, optional): initialize the cls token to be equal to the bias in the `nn.Linear`. Defaults to False.
            init_mask_as_bias (bool, optional): initialize the masking token to be equal to the bias in the `nn.Linear`. Defaults to False.
            init_cls_as_mask (bool, optional): initialize the cls token to be equal to the masking token. Defaults to False.
        """
        super().__init__()
        assert not (init_cls_as_bias and init_cls_as_mask)
        self.embedder = nn.Linear(1, embedding_dim)

        layers = []
        layers.append(nn.Linear(1, embedding_dim))
        layers.append(nn.LayerNorm(embedding_dim) if norm else nn.Identity())

        for _ in range(depth - 1):
            layers.append(nn.SiLU())
            layers.append(nn.Linear(embedding_dim, embedding_dim))
            layers.append(nn.LayerNorm(embedding_dim) if norm else nn.Identity())
        self.embedder = nn.Sequential(*layers)

        with torch.no_grad():
            bias = self.embedder(torch.tensor([[0.0]])).unsqueeze(0)

        if init_mask_as_bias:
            self.mask_embedding = nn.Parameter(bias)
        else:
            self.mask_embedding = nn.Parameter(torch.empty(embedding_dim))
            nn.init.uniform_(self.mask_embedding, -1, 1)

        if cls:
            if init_cls_as_bias:
                init = bias
            elif init_cls_as_mask:
                init = self.mask_embedding.data.clone()
            else:
                init = None
            self.cls = CLS(embedding_dim, cls_axis=cls_axis, init=init)
        else:
            self.cls = nn.Identity()

    def forward(self, x):
        out = self.embedder(x.unsqueeze(-1))
        out[torch.isnan(x)] = self.mask_embedding
        out = self.cls(out)
        return out


class BinaryEmbedding(nn.Module):
    def __init__(
        self,
        embedding_dim,
        lambda_fn=lambda x: x == 0,
        cls=False,
        cls_axis=-2,
        init_cls_as_mask=False,
    ):
        """Embedding module that binarizes input data and then embeds them using DiscreteEmbedding. Supports masking tokens as NaNs
        For efficiency, consider tokenizing data before training using this module and using DiscreteEmbedding in the actual model.

        Args:
            embedding_dim (int): number of hidden dimensions of the embeddings
            lambda_fn (Callable, optional): function that binarizes data. Defaults to lambdax:x==0.
        """
        super().__init__()
        self.embedder = DiscreteEmbedding(
            2,
            embedding_dim,
            cls=cls,
            cls_axis=cls_axis,
            init_cls_as_mask=init_cls_as_mask,
        )
        self.lambda_fn = lambda_fn

    def forward(self, x):
        nans = torch.isnan(x)
        x = self.lambda_fn(x).float()
        x[nans] = torch.nan
        return self.embedder(x)


class BinEmbedding(nn.Module):
    def __init__(
        self,
        embedding_dim,
        bins,
        left_closed=True,
        cls=False,
        cls_axis=-2,
        init_cls_as_mask=False,
    ):
        """Embedding module that bins input data and then embeds them using DiscreteEmbedding. Supports masking tokens as NaNs.
        For efficiency, consider tokenizing data before training using this module and using DiscreteEmbedding in the actual model.

        Args:
            embedding_dim (int): number of hidden dimensions of the embeddings
            bins (torch.Tensor): 1-D tensor that delineates bins. Should be monotonically increasing.
            left_closed (bool): Whether the bins should start including the lower bound or not. Defaults to True
        """
        super().__init__()
        self.embedder = DiscreteEmbedding(
            len(bins),
            embedding_dim,
            cls=cls,
            cls_axis=cls_axis,
            init_cls_as_mask=init_cls_as_mask,
        )
        self.left_closed = left_closed
        self.bins = bins

    def forward(self, x):
        nans = torch.isnan(x)
        bins = self.bin(x).float()
        bins[nans] = torch.nan
        return self.embedder(bins)

    def bin(self, x):
        diff = x[..., None] - self.bins
        z = diff >= 0 if self.left_closed else diff > 0
        z = torch.cat([z, torch.zeros_like(z)[..., [-1]]], dim=-1)
        return torch.clamp(z.int().argmin(2) - 1, min=0)


class CLS(nn.Module):
    def __init__(self, dim, cls_axis=-2, init=None):
        """Module for adding a CLS token to an input

        Args:
            dim (int): number of hidden dimensions of the embeddings
            cls_axis (int, optional): which axis in the input to add cls token to. Defaults to -1.
            init (int, optional): initialize the weights of the CLS token as ... Defaults to None, for uniform [-1, 1] initialization.
        """
        super().__init__()
        if init is not None:
            self.cls_embedding = nn.Parameter(init)
        else:
            self.cls_embedding = nn.Parameter(torch.empty(dim))
            nn.init.uniform_(self.cls_embedding, -1, 1)

        self.cls_axis = cls_axis

    def forward(self, x):
        shp = list(x.shape)
        shp[self.cls_axis] = 1
        x = torch.cat([self.cls_embedding.expand(*shp), x], self.cls_axis)
        return x
